class RelationInQuery:

    def __init__(self, name='', variables={},constants = {},wildcards = [], negated = False):
        self.name = name # The name of the relation
        self.variables = variables # {Variable -> [Position]}
        self.constants = constants # {Constant -> [Position]}
        self.negated = negated # Is it negated
        self.wildcards = wildcards # [position]

    def __init__(self, name='', elements = [],negated = False):
        self.name = name
        self.negated = negated # Is it negated

        self.variables = {} # {Variable -> [Position]}
        self.constants = {} # {Constant -> [Position]}
        self.wildcards = [] # Is it negated

        for element in elements:
            if element.is_variable():
                self.variables[element] = []
            if element.is_constant():
                self.constants[element] = []

        # Build the variables, constants dictionaries and the wildcard list.
        for i in range(0,len(elements)):
            element = elements[i]
            if element.is_variable():
                self.variables[element].append(i)
            if element.is_constant():
                self.constants[element].append(i)
            if element.is_wildcard():
                self.wildcards.append(i)        

    def __repr__(self):
        list_of_columns = list(range(0,len([y for x in self.variables.values() for y in x]) + \
                                  len([y for x in self.constants.values() for y in x]) + \
                                  len(self.wildcards)))

        for variable in self.variables:
            for position in self.variables[variable]:
                list_of_columns[position] = str(variable)
        for constant in self.constants:
            for position in self.constants[constant]:
                 list_of_columns[position] = str(constant)
        for position in self.wildcards:
            list_of_columns[position] = '_'

        neg_str = '!' if self.negated else ''

        return neg_str + self.name + '('+','.join(list_of_columns)+')'

    def get_name(self):
        return self.name

    def set_name(self,name):
        self.name = name

    def get_variables(self):
        return self.variables

    def get_constants(self):
        return self.constants

    def get_wildcards(self):
        return self.wildcards

    def is_negated(self):
        return self.negated

    def get_ordered_element_list(self):

        order_dict = {}

        variables = self.get_variables()
        for var in variables:
            for pos in variables[var]:
                order_dict[pos] = var

        constants = self.get_constants()
        for const in constants:
            for pos in constants[const]:
                order_dict[pos] = const


        for pos in self.get_wildcards():
            order_dict[pos] = '_'

        return [order_dict[index] for index in range(0,len(order_dict))]



class Element:
    def __eq__(self, other):
        return (isinstance(other, self.__class__)
            and self.__dict__ == other.__dict__)
    def is_variable(self):
        return isinstance(self,Variable)
    def is_constant(self):
        return isinstance(self,Constant)
    def is_wildcard(self):
        return isinstance(self,Wildcard)

class Variable(Element):
    def __init__(self,name = ''):
        self.name = name
    def __hash__(self):
        return hash(self.name)
    def __repr__(self):
        return self.name

class Constant(Element):
    def __init__(self,value = ''):
        self.value = value
    def __hash__(self):
        return hash(self.value)
    def __repr__(self):
        return self.value

class Wildcard(Element):
    def __hash__(self):
        return hash('_')
    def __repr__(self):
        return '_'




__author__ = 'caioseguin'

from .RuleHandler import RuleHandler

# This class acts as the middle man between the Datalog parser and the SQL translator.
# It transforms the parser's output into a list of conjunctiveQueries and a list of disjunctiveQueries.

class ASTBuilder:
    # Each element of the list program_ast consists in a list. Each sub-list start with a conjunctiveQuery or a
    # disjunctiveQuery Object, followed by a number of relationInQuery Objects. A new sub-list is appended every time a
    # new line from the parsed datalog program is handled.
    # The program_ast is going to be passed along statement by statement.
    # The history_ast is the whole history of datalog programs handled by the ASTFactory.

    # The RuleHandler is responsible of translating every rule into the ast form

    history_ast = []
    program_ast = []
    rule_handler = RuleHandler()

    def __init__(self, script_input_flag = 1):
        self.script_input_flag = script_input_flag
        history_ast = []
        program_ast = []
        rule_handler = RuleHandler()


    def buildAST(self, parsed_datalog_program):

        assert isinstance(parsed_datalog_program, list)

        self.program_ast = []

        for code_line in parsed_datalog_program:
            x = self.handleStatement(code_line)
            self.program_ast.append(x)

        self.history_ast.append(self.program_ast)

        return self.program_ast

    def handleCodeLine(self, code_line):

        assert isinstance(code_line, list)

        code_line_ast = []

        for statement in code_line:
            statement_ast = self.handleStatement(statement)
            code_line_ast.append(statement_ast)

        return code_line_ast

    def handleStatement(self, statement):

        assert isinstance(statement, list)

        if self.script_input_flag:

            if self.rule_handler.isRule(statement):
                statement_ast = self.handleRule(statement)
            else:
                statement_ast = self.handleFact(statement)

        else:
            raise Exception("?.handleQuery(statement) not implemented yet")

        return statement_ast

    def handleRule(self, statement):

        assert isinstance(statement, list)

        if self.rule_handler.isDisjunctiveRule(statement):
            raise Exception("self.ruleHandler.handleDisjunctiveRule(statement) not implemented yet")
        else:
            return self.rule_handler.handleConjunctiveRule(statement)

    def handleFact(self, statement):
        assert isinstance(statement, list)
        return self.rule_handler.handleFact(statement)




from .translateDatalogToSQL import translateDatalogToSql

from ..evaluate_query import *

__author__ = 'caioseguin'

def main():
    datatog_statements = []
    datatog_statements.append('users(X,Y):- tuser(X,Y).')
    datatog_statements.append('users(Y,L):-tuser(_,Y,_,L).')
    datatog_statements.append("users(Y):-tuser(_,Y,_,L), L='Lyon'.")
    #datatog_statements.append('q(X):-not(s(Y)),X=Y,Y=2.')
    #datatog_statements.append('q(X):- s(X), X = 2.')
    #datatog_statements.append('r(X,Y),!s(Z), Y=Z.')
    #datatog_statements.append('r(X,Y),!s(Z).')
    #datatog_statements.append('a(X,Y) :- r(X,A), s(A,Y).')
    #datatog_statements.append('a(X,2) :- r(X,A), s(A,_).')
    #datatog_statements.append('r(_,2).')
    #datatog_statements.append('r(_,A), s(A,_)..')
    #datatog_statements.append('r(X,Y) :- s(X), s(Y), X>Y.')
    #datatog_statements.append('r(X,Y) :- s(X), Y>2')
    #datatog_statements.append('r(X) :- s(X), X<2')
    #datatog_statements.append('a(X,Y):-s(X,Z),s(Y,Z),X>Y.')
    #datatog_statements.append('s(X) :- X = 2, 3<X.')
    #datatog_statements.append('r(_,2), X = 2.')
    #datatog_statements.append('a(X,Y) :- s(X), sa(Y,T), t(X), u(Y), X>Y.')
    #datatog_statements.append("q(X,Y):- s(X,Y). a(X,Y):- b(X,Y).")

    sql_list=[]
    for d_query in datatog_statements:
        #print(translateDatalogToSql(d_query))
        s = translateDatalogToSql(d_query)
        sql_list.append(s[0])

    for sqlS in sql_list:
        evaluateQuery().evaluate(sqlS)


main()



__author__ = 'nishara'



from sqlalchemy import *
from tabulate import tabulate
#TODO: allow db choice
engine = create_engine('postgresql://yadi_user:@localhost:5432/cinema', echo=True)


class evaluateQuery():
    def evaluate(self,query_to_evaluate):
        r = engine.execute(query_to_evaluate)
        fetched_results = r.fetchall()
        print (tabulate(fetched_results, tablefmt="grid"))



#from .dataStructures.query import *
#from .dataStructures.relation import *
#from .dataStructures.element import *
#from .dataStructures.constraint import *
#from .queryExecutor.queryExecutor import *
#
#'''
#r = RelationInQuery('', {Variable():,}, {Constant():,},[],False)
#q = ConjunctiveQuery([],[],[Variable(),])
#
#class RelationInQuery:
#
#    def __init__(self, name='', variables={},constants = {},wildcards = [], is_negated = False):
#        self.name = name # The name of the relation
#        self.variables = variables # {Variable -> Position}
#        self.constants = constants # {Constant -> Position}
#        self.negated = False # Is it negated
#        self.wildcards = wildcards # [position]
#
#
#class Query:
#    def __init__(self, relations = [], constraints = [], head_variables = [] ):
#        self.head_variables = head_variables
#        self.relations = relations# [RelationInQuery].
#        self.constraints = constraints # Explicit constraints of the form Element COMP Element type.
#                                       # [(Element_1, Element_2, Comparison_operator)]
#                                       '''
#
#
#import copy
#
#
#
## SAFETY TESTS
## Q(X,Y):- S(X) Not Safe
#
#r = RelationInQuery('S', {Variable('X'):[0]}, {},[],False)
#q = ConjunctiveQuery([r],[],RelationInQuery('Q',{Variable('X'):[0],Variable('Y'):[1]}))
#bla = QueryToAlchemyStatement(q)
#print ('Test 1')
#print ('Correct: False')
#try:
#    print (bla.check_is_it_safe())
#except Exception as e:
#    print (e)
#
#print ('-------------------')
#print ('')
#
## Q(X):- S(X) Safe
#
#r = RelationInQuery('S', {Variable('X'):[0]}, {},[],False)
#q = ConjunctiveQuery([r],[],RelationInQuery('Q',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print ('Test 2')
#print ('Correct: True')
#try:
#    print bla.check_is_it_safe()
#except Exception as e:
#    print e
#print ('-------------------')
#print ('')
#
## Q(X):- S(X), !T(U) Not safe
#
#s = RelationInQuery('S', {Variable('X'):[0]}, {},[],False)
#t = RelationInQuery('T', {Variable('U'):[0]}, {},[],True)
#q = ConjunctiveQuery([s,t],[],RelationInQuery('Q',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print ('Test 3')
#print ('Correct: False')
#try:
#    print (bla.check_is_it_safe())
#except Exception as e:
#    print (e)
#print ('-------------------')
#print ('')
#
## Q(X):- S(X,U), !T(U) Safe
#
#s = RelationInQuery('S', {Variable('X'):[0],Variable('U'):[1]}, {},[],False)
#t = RelationInQuery('T', {Variable('U'):[0]}, {},[],True)
#q = ConjunctiveQuery([s,t],[],RelationInQuery('Q',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 4'
#print 'Correct: True'
#try:
#    print bla.check_is_it_safe()
#except Exception as e:
#    print e
#print '-------------------'
#print ''
#
## Q(X):- S(X), !T(U), U = 2 Safe
#
#s = RelationInQuery('S', {Variable('X'):[0],Variable('U'):[1]}, {},[],False)
#t = RelationInQuery('T', {Variable('U'):[0]}, {},[],True)
#q = ConjunctiveQuery([s,t],[Constraint(Variable('U'), Constant('2'), '==')],RelationInQuery('Q',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 5'
#print 'Correct: True'
#try:
#    print bla.check_is_it_safe()
#except Exception as e:
#    print e
#print '-------------------'
#print ''
#
## Q(X):- S(X), !T(U), U < 2 Not safe
#
#s = RelationInQuery('S', {Variable('X'):[0]}, {},[],False)
#t = RelationInQuery('T', {Variable('U'):[0]}, {},[],True)
#q = ConjunctiveQuery([s,t],[Constraint(Constant('2'),Variable('U'), '<')],RelationInQuery('Q',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 6'
#print 'Correct: False'
#try:
#    print bla.check_is_it_safe()
#except Exception as e:
#    print e
#print '-------------------'
#print ''
#
## Q(X):- S(X), !T(U), U=X Safe
#
#s = RelationInQuery('S', {Variable('X'):[0]}, {},[],False)
#t = RelationInQuery('T', {Variable('U'):[0]}, {},[],True)
#q = ConjunctiveQuery([s,t],[Constraint(Variable('U'), Variable('X'), '=')],RelationInQuery('Q',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 7'
#print 'Correct: True'
#try:
#    bla.query= bla.preProcessQuery(bla.query)
#    print bla.check_is_it_safe()
#except Exception as e:
#    print e
#
#print '-------------------'
#print ''
#
## Q(X):- S(X), !T(U), U=X Safe
#
#s = RelationInQuery('S', {Variable('X'):[0]}, {},[],False)
#t = RelationInQuery('T', {Variable('U'):[0]}, {},[],True)
#q = ConjunctiveQuery([s,t],[Constraint(Variable('U'), Variable('X'), '=')],RelationInQuery('Q',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 8'
#print 'Correct: True'
#try:
#    bla.query= bla.preProcessQuery(bla.query)
#    print bla.check_is_it_safe()
#except Exception as e:
#    print e
#
#print '-------------------'
#print ''
#
## Q(X):- S(X), Y<3 Not safe
#
#s = RelationInQuery('S', {Variable('X'):[0]}, {},[],False)
#
#q = ConjunctiveQuery([s],[Constraint(Variable('Y'), Constant('3'), '<')],RelationInQuery('Q',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 9'
#print 'Correct: False'
#try:
#    bla.query= bla.preProcessQuery(bla.query)
#    print bla.check_is_it_safe()
#except NotSafeException as e:
#    print e
#
#print '-------------------'
#print ''
#
## Q(X):- S(X,Y), Y<3 Safe
#
#s = RelationInQuery('S', {Variable('X'):[0],Variable('Y'):[1]}, {},[],False)
#
#q = ConjunctiveQuery([s],[Constraint(Variable('Y'), Constant('3'), '<')],RelationInQuery('Q',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 10'
#print 'Correct: True'
#try:
#    bla.query= bla.preProcessQuery(bla.query)
#    print bla.check_is_it_safe()
#except Exception as e:
#    print e
#
#print '-------------------'
#print ''
#
## Q(X):- S(X), !T(Y), X=Y
#
#s = RelationInQuery('S', {Variable('X'):[0]}, {},[],False)
#t = RelationInQuery('T', {Variable('Y'):[1]}, {},[],True)
#
#q = ConjunctiveQuery([s,t],[Constraint(Variable('X'), Variable('Y'), '=')],RelationInQuery('Q',{Variable('X'):[0]}))
#
#bla = QueryToAlchemyStatement(q)
#print 'Test 11'
#print 'Correct: True'
#try:
#    bla.query= bla.preProcessQuery(bla.query)
#    print bla.check_is_it_safe()
#except Exception as e:
#    print e
#
#print '-------------------'
#print ''
#
## Q(X,Z):- S(X,Y), !T(A,Z), Z = Y
#
#s = RelationInQuery('S', {Variable('X'):[0],Variable('Y'):[1]}, {},[],False)
#t = RelationInQuery('T', {Variable('A'):[0],Variable('Z'):[1]}, {},[],True)
#
#q = ConjunctiveQuery([s,t],[Constraint(Variable('Y'), Variable('Z'), '=')],RelationInQuery('Q',{Variable('X'):[0],Variable('Z'):[1]}))
#
#bla = QueryToAlchemyStatement(q)
#print 'Test 12'
#print 'Correct: False'
#try:
#    bla.query= bla.preProcessQuery(bla.query)
#    print bla.check_is_it_safe()
#except Exception as e:
#    print e
#
#print '-------------------'
#print ''
#
#
## EQUALITY CONSTRAINTS REDUCTION TESTS:
#
## Q(X):- S(X), X = 2
#r = RelationInQuery('S', {Variable('X'):[0]}, {},[],False)
#q = ConjunctiveQuery([r],[Constraint(Variable('X'), Constant('2'), '=')],RelationInQuery('S',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 1'
#print 'Original: '
#print q
#print ''
#print bla.reduce_equality_constraints(q)
#print '-------------------'
#print ''
#
## Q(X):- S(X, Y), Y = 2
#r = RelationInQuery('S', {Variable('X'):[0], Variable('Y'):[1]}, {},[],False)
#q = ConjunctiveQuery([r],[Constraint(Variable('Y'), Constant('2'), '=')],RelationInQuery('S',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 2'
#print 'Original: '
#print q
#print ''
#print bla.reduce_equality_constraints(q)
#print '-------------------'
#print ''
#
## Q(X):- S(X, Y, Z), Y = 2, Z = Y
#r = RelationInQuery('S', {Variable('X'):[0], Variable('Y'):[1], Variable('Z'):[2]}, {},[],False)
#q = ConjunctiveQuery([r],[Constraint(Variable('Y'), Constant('2'), '='), Constraint(Variable('Z'), Variable('Y'), '=')],RelationInQuery('S',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 3'
#print 'Original: '
#print q
#print ''
#print bla.reduce_equality_constraints(q)
#print '-------------------'
#print ''
#
##  Q(X) := not S(Y),  X = 2, X = Y,
#r = RelationInQuery('S', {Variable('Y'):[0]}, {},[],True)
#q = ConjunctiveQuery([r],[Constraint(Variable('X'), Constant('2'), '='), Constraint(Variable('X'), Variable('Y'), '=')],RelationInQuery('S',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 4'
#print 'Original: '
#print q
#print ''
#print bla.reduce_equality_constraints(q)
#print '-------------------'
#print ''
#
##  Q(X) := not S(Y), X = Y, X = 2
#r = RelationInQuery('S', {Variable('Y'):[0]}, {},[],True)
#q = ConjunctiveQuery([r],[Constraint(Variable('X'), Variable('Y'), '='),Constraint(Variable('X'), Constant('2'), '=')],RelationInQuery('S',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 5'
#print 'Original: '
#print q
#print ''
#print bla.reduce_equality_constraints(q)
#print '-------------------'
#print ''
#
## Q(X) := S(X), X = Y
#r = RelationInQuery('S', {Variable('X'):[0]}, {},[],False)
#q = ConjunctiveQuery([r],[Constraint(Variable('X'), Variable('Y'), '=')],RelationInQuery('S',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 6'
#print 'Original: '
#print q
#print ''
#print bla.reduce_equality_constraints(q)
#print '-------------------'
#print ''
#
## Q(X) := S(Y), X = Y
#r = RelationInQuery('S', {Variable('Y'):[0]}, {},[],False)
#q = ConjunctiveQuery([r],[Constraint(Variable('Y'), Variable('X'), '=')],RelationInQuery('S',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 7'
#print 'Original: '
#print q
#print ''
#print bla.reduce_equality_constraints(q)
#print '-------------------'
#print ''
#
## Q(X) := S(X), P = Y
#
#r = RelationInQuery('S', {Variable('X'):[0]}, {},[],False)
#q = ConjunctiveQuery([r],[Constraint(Variable('Y'), Variable('P'), '=')],RelationInQuery('S',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 8'
#print 'Original: '
#print q
#print ''
#try:
#    print bla.reduce_equality_constraints(q)
#except Exception as e:
#    print e
#print '-------------------'
#print ''
#
#
#s = RelationInQuery('S', {Variable('X'):[0],Variable('Y'):[1]}, {},[],False)
#r = RelationInQuery('R', {Variable('T'):[0],Variable('Z'):[1]}, {},[],False)
#q = ConjunctiveQuery([s,r],[Constraint(Variable('X'), Variable('T'), '='),Constraint(Variable('T'), Variable('X'), '='),Constraint(Variable('Y'), Variable('Z'), '='),Constraint(Variable('T'), Variable('Z'), '='), Constraint(Variable('Z'),Constant('2'), '='),Constraint(Variable('P'), Variable('Q'), '='),Constraint(Variable('P'), Variable('T'), '=')],RelationInQuery('S',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 9'
#print 'Original: '
#print q
#print ''
#print bla.reduce_equality_constraints(q)
#print '-------------------'
#print ''
#
#s = RelationInQuery('S', {Variable('X'):[0,1]}, {},[],False)
#r = RelationInQuery('R', {Variable('T'):[0,1]}, {},[],False)
#q = ConjunctiveQuery([s,r],[Constraint(Variable('X'), Variable('T'), '='),Constraint(Variable('T'), Variable('X'), '='),Constraint(Variable('Y'), Variable('Z'), '='),Constraint(Variable('T'), Variable('Z'), '='), Constraint(Variable('Z'),Constant('2'), '='),Constraint(Variable('P'), Variable('Q'), '='),Constraint(Variable('P'), Variable('T'), '=')],RelationInQuery('S',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 10'
#print 'Original: '
#print q
#print ''
#print bla.reduce_equality_constraints(q)
#print '-------------------'
#print ''
#
## R(X) := S(X), 3=2, X>2
#r = RelationInQuery('S', {Variable('X'):[0]}, {},[],True)
#q = ConjunctiveQuery([r],[Constraint(Constant('3'), Constant('2'), '='),Constraint(Variable('X'), Constant('2'), '>')],RelationInQuery('S',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 11'
#print 'Original: '
#print q
#print ''
#print bla.reduce_equality_constraints(q)
#print '-------------------'
#print ''
#
#s = RelationInQuery('S', {Variable('X'):[0],Variable('P'):[1]}, {},[],False)
#r = RelationInQuery('R', {Variable('T'):[0],Variable('Z'):[1]}, {},[],False)
#q = ConjunctiveQuery([s,r],[Constraint(Variable('X'), Variable('T'), '='),Constraint(Variable('T'), Variable('X'), '='),Constraint(Variable('Y'), Variable('Z'), '='),Constraint(Variable('T'), Variable('Z'), '='), Constraint(Variable('Z'),Constant('2'), '='),Constraint(Variable('P'), Variable('Q'), '='),Constraint(Variable('Y'), Variable('T'), '=')],RelationInQuery('S',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 12'
#print 'Original: '
#print q
#print ''
#print bla.reduce_equality_constraints(q)
#print '-------------------'
#print ''
#
#
#s = RelationInQuery('S', {Variable('X'):[0],Variable('P'):[1]}, {},[],False)
#r = RelationInQuery('R', {Variable('T'):[0],Variable('Z'):[1]}, {},[],False)
#q = ConjunctiveQuery([s,r],[Constraint(Variable('X'), Variable('T'), '='),Constraint(Variable('T'), Variable('X'), '='),Constraint(Variable('Y'), Variable('Z'), '='),Constraint(Variable('T'), Variable('Z'), '='), Constraint(Variable('Z'),Constant('2'), '='),Constraint(Variable('P'), Variable('Q'), '='),Constraint(Variable('Q'), Constant('5'), '=')],RelationInQuery('S',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 13'
#
#print 'Original: '
#print q
#print ''
#print bla.reduce_equality_constraints(q)
#print '-------------------'
#print ''
#
#
## R(X) := S(X), 3=X, X>2
#r = RelationInQuery('S', {Variable('X'):[0]}, {},[],True)
#q = ConjunctiveQuery([r],[Constraint(Constant('3'), Variable('X'), '='),Constraint(Variable('X'), Constant('2'), '>')],RelationInQuery('S',{Variable('X'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 14'
#print 'Original: '
#print q
#print ''
#print bla.reduce_equality_constraints(q)
#print '-------------------'
#print ''
#
## R(Y) := S(X), Y=X, X=2 Y =4
#r = RelationInQuery('S', {Variable('Y'):[0]}, {},[],True)
#q = ConjunctiveQuery([r],[Constraint(Variable('Y'), Variable('X'), '='),Constraint(Variable('X'), Constant('2'), '='),Constraint(Variable('Y'), Constant('4'), '=')],RelationInQuery('S',{Variable('Y'):[0]}))
#bla = QueryToAlchemyStatement(q)
#print 'Test 15'
#print 'Original: '
#print q
#print ''
#print bla.reduce_equality_constraints(q)
#print '-------------------'
#print ''
#
## R(X,Y,Z,_,2) := S(X), Y=X, X=2
#r = RelationInQuery('S', {Variable('Y'):[0]}, {},[],True)
#q = ConjunctiveQuery([r],[Constraint(Variable('Y'), Variable('X'), '='),Constraint(Variable('X'), Constant('2'), '='),Constraint(Variable('Y'), Constant('4'), '=')],RelationInQuery('R',{Variable('X'):[0],Variable('Y'):[1],Variable('Z'):[2]},{Constant('2'):[4]},[3]))
#bla = QueryToAlchemyStatement(q)
#print 'Test 16'
#print 'Original: '
#print q
#print ''
#print bla.reduce_equality_constraints(q)
#print '-------------------'
#print ''
#



from ..dataStructures.query import *
from .exceptions import *

class SafetyChecker():
    def check_for_safety(self,query):
        if isinstance(query, ConjunctiveQuery):
            sq = ConjunctiveQuerySafetyChecker()
        if isinstance(query, DisjunctiveQuery):
            sq = DisjunctiveQuerySafetyChecker()
        return sq.check_for_safety(query)

class ConjunctiveQuerySafetyChecker():

    def check_for_safety(self,query):
        var_dict = query.get_var_dict()
        safe_variables = self.get_safe_variables(query,var_dict)
        return (self.check_head(safe_variables,query) and
               self.check_negated_goals(safe_variables,query) and
               self.check_non_equality_explicit_constraints(safe_variables,query))

    def check_head(self,safe_variables,query):

        for variable in list(query.head_relation.get_variables()):
            if not (variable in safe_variables):
                raise NotSafeException('Query not safe because '+ variable.name + ' occurs in the head and not in a positive goal')
                return False
        return True

    def check_negated_goals(self,safe_variables,query):

        # Create a list of variables which occur in negated goals.
        variables_in_negated_goals = [y for x in query.relations for y in list(x.variables) if x.is_negated()]

        # And check them:
        for variable in variables_in_negated_goals:
            if not (variable in safe_variables):
                raise NotSafeException('Query not safe because ' + variable.name + ' from a negated goal does not occur in a positive goal')
                return False

        return True


    def check_non_equality_explicit_constraints(self,safe_variables,query):

        # Checking variables which occur in explicit constraints with non equality operators

        # Create a list of variables which occur in explicit constraints with non equality operators
        variables_in_constraints_with_non_equality_operators = [y for x in query.constraints \
                                                                    for y in [x.get_left_side(),x.get_right_side()] \
                                                                        if y.is_variable() and not x.is_equality_constraint()]

        for variable in variables_in_constraints_with_non_equality_operators:
            if not (variable in safe_variables):
                raise NotSafeException('Query not safe because ' + variable.name + ' from a non_equality comparison does not occur in a positive goal')
                return False

        return True

    def get_safe_variables(self,query,var_dict):
        # Create a list of safe variables: variables bounded to constants or that occur in a positive goal

        variables_bounded_to_constants = []

        for constraint in query.constraints:
            if (constraint.get_left_side().is_variable() and
                constraint.get_right_side().is_constant() and
                constraint.is_equality_constraint()):
                    variables_bounded_to_constants.append(constraint.get_left_side())

        safe_variables = list(var_dict) + variables_bounded_to_constants

        return safe_variables





from .TranslationWrapper.translateDatalogToSQL import translateDatalogToSql
from .evaluate_query import *

def start():
    while True:
        prompt = 'yadi> '
        line = input ('yadi> ')
        sql_queries = translateDatalogToSql(line)
        for s in sql_queries:
            try:
                evaluateQuery().evaluate(s)
            except Exception as e:
                print(e)
start()



from ..dataStructures.query import *

class SQLGenerator():
    def get_SQL_code(self,query,old_query):
        if isinstance(query, ConjunctiveQuery):
            sql_gen = ConjunctiveQuerySQLGenerator()
        if isinstance(query, DisjunctiveQuery):
            sql_gen = DisjunctiveQuerySQLGenerator()
        return sql_gen.get_SQL_code(query,old_query)

class DisjunctiveQuerySQLGenerator():

    def get_SQL_code(self,query,old_query):
        sql_gen = SQLGenerator()
        ' UNION '.join([sql_gen.get_SQL_code(q) for q in query.get_queries()])

class ConjunctiveQuerySQLGenerator():

    def get_SQL_code(self,query,old_query,pretty_print=False):
        aliases = self.create_table_aliases(query)
        var_dict = query.get_var_dict()

        select_clause = self.get_select_columns(query.get_head_relation(), query.get_constraints(), var_dict, old_query.get_head_relation())
        from_clause = self.get_from_relations(query,aliases)
        implicit_constraints = self.get_implicit_constraints(query.get_relations())
        join_constraints = self.get_join_constraints(var_dict)
        explicit_constraints = self.get_explicit_constraints(query.get_constraints(),var_dict)
        negated_queries = self.get_negated_queries(query.get_relations(),var_dict)
        where_separator = ' AND \n\t' if pretty_print else ' AND '
        where_clause = where_separator.join(implicit_constraints + join_constraints + explicit_constraints + negated_queries)
        if pretty_print:
            return 'SELECT \n\t' + ', \n\t'.join(select_clause) + \
               ('\nFROM \n\t' if len(from_clause) != 0 else '') + \
               ', \n\t'.join(from_clause) + \
               ('\nWHERE \n\t' if len(where_clause) != 0 else '') + where_clause + ';'
        else:
            return 'SELECT ' + ', '.join(select_clause) + \
               (' FROM ' if len(from_clause) != 0 else '') + \
               ', '.join(from_clause) + \
               (' WHERE ' if len(where_clause) != 0 else '') + where_clause + ';'



    def create_table_aliases(self,query):
        count = {}
        for relation in query.get_relations():
            if not (relation.get_name() in count):
                count[relation.get_name()] = 0
            else:
                count[relation.get_name()]+=1

        aliases = {}

        repeated_relations = [relation for relation in query.get_relations() if count[relation.get_name()]>0]
        non_repeated_relations = [relation for relation in query.get_relations() if count[relation.get_name()]==0]

        for relation in repeated_relations:
            old_name = relation.get_name()
            relation.set_name(old_name+str(count[relation.get_name()]))
            count[old_name]-=1
            aliases[relation.get_name()] = old_name + ' as ' + relation.get_name()

        for relation in non_repeated_relations:
            aliases[relation.get_name()] = relation.get_name()

        return aliases

    def mapVariableToRelationDotField(self,var,var_dict,pos = 0):
        relation = var_dict[var][pos][0]
        position = var_dict[var][pos][1]
        return relation.get_name() + '.'+ self.mapPositionToColumnName(relation,position)

    def get_from_relations(self, query,aliases):
        '''
        Returns list of the names of all positive relations in the query
        '''
        return [aliases[rel.get_name()] for rel in query.get_relations() if not rel.is_negated()]

    def mapPositionToColumnName(self,relation,position):
        return '_' + str(position)

    def get_select_columns(self,head_relation,constraints,var_dict, original_head_relation):
        column_list = []

        element_list = head_relation.get_ordered_element_list()

        if len(element_list) == 0:
            column_list = ["'TRUE'"]
        else:
            if original_head_relation is None:
                as_names = head_relation.get_ordered_element_list()
            else:
                as_names = original_head_relation.get_ordered_element_list()
            for i in range(0,len(element_list)):
                element = element_list[i]
                if element.is_variable():
                    if element in var_dict:
                    # It's in a positive goal.
                        column_list.append(self.mapVariableToRelationDotField(element, var_dict) + ' as ' + str(as_names[i]))
                    else:# It is a constant (it appears in a constraint).
                        const = str([x.get_right_side() for x in constraints if (x.get_left_side() == element and x.is_equality_constraint())][0])
                        column_list.append(const + ' as ' + str(as_names[i]))
                if element.is_constant():
                    column_list.append(str(element))
                if element.is_wildcard():
                    pass


        return column_list

    def get_implicit_constraints(self,relations):
        ''' An implicit constraint is one in which you are unifying a constant to a position of the relation
            E.G. R(X,2) specifies that R.2 == 2'''

        constraints = []
        for relation in [r for r in relations if not r.is_negated()]  :
            const_dict = relation.get_constants()
            for constant in const_dict:
                for position in const_dict[constant]:
                    constraints.append( \
                        relation.get_name() + '.' + self.mapPositionToColumnName(relation,position) + ' = ' + str(constant) \
                    )
        return constraints

    def get_explicit_constraints(self,constraints,var_dict):
        '''
        Returns SQL representation of an explicit constraint. An explicit
        constraint is one of the form Element COMP Element type, explicitly
        listed in the conjunctive query. E.G. R(X,Y), Y>2 specifies that R.2 > 2
        '''

        constraints_strings = []
        for constraint in constraints:
            ls = constraint.get_left_side()
            rs = constraint.get_right_side()

            if not (ls.is_variable() and constraint.is_equality_constraint()):
                ''' Every Var = Constant constraint where var is only in head. We don't want this type of constraints.
                    We assume that if there is a Var = Constraint, var occurs in the head. Otherwise it would be unified in the preprocessor.
                '''
                if ls.is_variable():
                    left_side = self.mapVariableToRelationDotField(ls, var_dict)
                elif ls.is_constant():
                    left_side = str(ls)

                if rs.is_variable():
                    right_side = self.mapVariableToRelationDotField(rs, var_dict)
                elif rs.is_constant():
                    right_side = str(rs)

                constraints_strings.append(
                    str(left_side) + ' ' + str(constraint.get_operator()) + ' ' + str(right_side)
                )

        return constraints_strings

    def get_join_constraints(self,var_dict):
        constraints = []

        for var in var_dict:
            length = len(var_dict[var])
            for i in range(1,length):
                constraints.append(
                    self.mapVariableToRelationDotField(var,var_dict,0) +
                    ' = ' +
                    self.mapVariableToRelationDotField(var,var_dict,i)
                )
        return constraints

    def get_negated_queries(self,relations,var_dict):
        negated_queries_sql = []
        for relation in [r for r in relations if r.is_negated()]:
            where_clauses = []
            elements = relation.get_ordered_element_list()

            for position in range(0,len(elements)):
                element = elements[position]
                var =  relation.get_name() + '.' + \
                       self.mapPositionToColumnName(relation, position)

                if element.is_variable():
                    where_clauses.append(
                        var + ' = ' +
                        self.mapVariableToRelationDotField(element, var_dict)
                    )
                elif element.is_constant():
                    where_clauses.append(var + ' = ' + str(element))

            negated_queries_sql.append(
                'NOT EXISTS (' +
                    'SELECT * FROM ' + relation.get_name() +
                    ' WHERE ' + ' AND '.join(where_clauses) +
                ')'
            )

        return negated_queries_sql





from .ASTBuilder import ASTBuilder
from ..dataStructures.constraint import Constraint
from ..dataStructures.element import Variable, Constant, Wildcard
from ..dataStructures.query import ConjunctiveQuery
from ..dataStructures.relation import RelationInQuery

__author__ = 'caioseguin'

# Function that compares the ASTFactory output with an expected input.
# It compares the two using the str() method.
# Unfortunately, it does not implement a type matching check.

def hardCodedTest():
    ast_builder = ASTBuilder()

    input_list = []
    expected_answer_list = []
    result_list = []

    # q(X,Y):- s(X)

    input = ([[['q', ['X', 'Y']], ':-', ['s', ['X']]]])
    expected_answer = ConjunctiveQuery([RelationInQuery('s', [Variable('X')], False)], [],
                                       RelationInQuery('q', [Variable('X'), Variable('Y')]))
    output = ast_builder.buildAST(input)[0]
    input_list.append(input)
    expected_answer_list.append(expected_answer)
    result_list.append(checkAnswer(output, expected_answer))

    # q(X):- s(X)

    input = ([[['q', ['X']], ':-', ['s', ['X']]]])
    expected_answer = ConjunctiveQuery([RelationInQuery('s', [Variable('X')], False)], [],
                                       RelationInQuery('q', [Variable('X')]))
    output = ast_builder.buildAST(input)[0]
    input_list.append(input)
    expected_answer_list.append(expected_answer)
    result_list.append(checkAnswer(output, expected_answer))

    # q(X):- s(X), not t(U)

    input = ([[['q', ['X']], ':-', ['s', ['X']], ['not', ['t', ['U']]]]])
    expected_answer = ConjunctiveQuery([RelationInQuery('s', [Variable('X')], False),
                                        RelationInQuery('t', [Variable('U')], True)], [],
                                       RelationInQuery('q', [Variable('X')]))
    output = ast_builder.buildAST(input)[0]
    input_list.append(input)
    expected_answer_list.append(expected_answer)
    result_list.append(checkAnswer(output, expected_answer))

    # q(X):- s(X,U), not t(U)

    input = ([[['q', ['X']], ':-', ['s', ['X', 'U']], ['not', ['t', ['U']]]]])
    expected_answer = ConjunctiveQuery([RelationInQuery('s', [Variable('X'), Variable('U')], False),
                                        RelationInQuery('t', [Variable('U')], True)], [],
                                       RelationInQuery('q', [Variable('X')]))
    output = ast_builder.buildAST(input)[0]
    input_list.append(input)
    expected_answer_list.append(expected_answer)
    result_list.append(checkAnswer(output, expected_answer))

    # q(X):- s(X), not t(U), U = 2

    input = ([[['q', ['X']], ':-', ['s', ['X']], ['not', ['t', ['U']]], ['U', '=', '2']]])
    expected_answer = ConjunctiveQuery([RelationInQuery('s', [Variable('X')], False),
                                        RelationInQuery('t', [Variable('U')], True)],
                                       [Constraint(Variable('U'), Constant('2'), '=')],
                                       RelationInQuery('q', [Variable('X')]))
    output = ast_builder.buildAST(input)[0]
    input_list.append(input)
    expected_answer_list.append(expected_answer)
    result_list.append(checkAnswer(output, expected_answer))

    # q(X):- s(X), not t(U), u < 2

    input = ([[['q', ['X']], ':-', ['s', ['X']], ['not', ['t', ['U']]], ['U', '<', '2']]])
    expected_answer = ConjunctiveQuery([RelationInQuery('s', [Variable('X')], False),
                                        RelationInQuery('t', [Variable('U')], True)],
                                       [Constraint(Variable('U'), Constant('2'), '<')],
                                       RelationInQuery('q', [Variable('X')]))
    output = ast_builder.buildAST(input)[0]
    input_list.append(input)
    expected_answer_list.append(expected_answer)
    result_list.append(checkAnswer(output, expected_answer))

    # q(X):- s(X), not t(U), U = X

    input = ([[['q', ['X']], ':-', ['s', ['X']], ['not', ['t', ['U']]], ['U', '=', 'X']]])
    expected_answer = ConjunctiveQuery([RelationInQuery('s', [Variable('X')], False),
                                        RelationInQuery('t', [Variable('U')], True)],
                                       [Constraint(Variable('U'), Variable('X'), '=')],
                                       RelationInQuery('q', [Variable('X')]))
    output = ast_builder.buildAST(input)[0]
    input_list.append(input)
    expected_answer_list.append(expected_answer)
    result_list.append(checkAnswer(output, expected_answer))

    # q(X,Y):- s(X,Y), y < 3

    input = ([[['q', ['X', 'Y']], ':-', ['s', ['X', 'Y']], ['Y', '<', '3']]])
    expected_answer = ConjunctiveQuery([RelationInQuery('s', [Variable('X'), Variable('Y')], False)],
                                       [Constraint(Variable('Y'), Variable('3'), '<')],
                                       RelationInQuery('q', [Variable('X'), Variable('Y')]))
    output = ast_builder.buildAST(input)[0]
    input_list.append(input)
    expected_answer_list.append(expected_answer)
    result_list.append(checkAnswer(output, expected_answer))

    # q(X):- s(X), not t(Y), X = Y

    input = ([[['q', ['X']], ':-', ['s', ['X']], ['not', ['t', ['Y']]], ['X', '=', 'Y']]])
    expected_answer = ConjunctiveQuery([RelationInQuery('s', [Variable('X')], False),
                                        RelationInQuery('t', [Variable('Y')], True)],
                                       [Constraint(Variable('X'), Variable('Y'), '=')],
                                       RelationInQuery('q', [Variable('X')]))
    output = ast_builder.buildAST(input)[0]
    input_list.append(input)
    expected_answer_list.append(expected_answer)
    result_list.append(checkAnswer(output, expected_answer))

    # q(X,Z):- s(X,Y), not t(A,Z), Z = Y

    input = ([[['q', ['X', 'Z']], ':-', ['s', ['X', 'Y']], ['not', ['t', ['A', 'Z']]], ['Z', '=', 'Y']]])
    expected_answer = ConjunctiveQuery([RelationInQuery('s', [Variable('X'), Variable('Y')], False),
                                        RelationInQuery('t', [Variable('A'), Variable('Z')], True)],
                                       [Constraint(Variable('Z'), Variable('Y'), '=')],
                                       RelationInQuery('q', [Variable('X'), Variable('Z')]))
    output = ast_builder.buildAST(input)[0]
    input_list.append(input)
    expected_answer_list.append(expected_answer)
    result_list.append(checkAnswer(output, expected_answer))

    # q(X):- s(X, Y, Z), Y = 2, Z = Y

    input = ([[['q', ['X']], ':-', ['s', ['X', 'Y', 'Z']], ['Y', '=', '2'], ['Z', '=', 'Y']]])
    expected_answer = ConjunctiveQuery([RelationInQuery('s', [Variable('X'), Variable('Y'), Variable('Z')], False)],
                                       [Constraint(Variable('Y'), Constant('2'), '='),
                                        Constraint(Variable('Z'), Variable('Y'), '=')],
                                       RelationInQuery('q', [Variable('X')]))
    output = ast_builder.buildAST(input)[0]
    input_list.append(input)
    expected_answer_list.append(expected_answer)
    result_list.append(checkAnswer(output, expected_answer))

    # r(X,'Nantes') :- r(X)

    input = ([[['r', ['X', "'Nantes'"]], ':-', ['r', ['X']]]])
    expected_answer = ConjunctiveQuery([RelationInQuery('r', [Variable('X')], False)],[],
                                       RelationInQuery('r', [Variable('X'), Constant("'Nantes'")]))
    output = ast_builder.buildAST(input)[0]
    input_list.append(input)
    expected_answer_list.append(expected_answer)
    result_list.append(checkAnswer(output, expected_answer))

    # r(X,Y,Z,_,2) :- s(X), Y=X, X=2

    input = ([[['r', ['X', 'Y', 'Z', '_', '2']], ':-', ['s', ['X']], ['Y', '=', 'X'], ['X', '=', '2']]])
    expected_answer = ConjunctiveQuery([RelationInQuery('s', [Variable('X')], False)],
                                       [Constraint(Variable('Y'), Variable('X'), '='),
                                        Constraint(Variable('X'), Variable('2'), '=')],
                                       RelationInQuery('r', [Variable('X'), Variable('Y'), Variable('Z'),
                                                             Wildcard(), Constant('2')]))
    output = ast_builder.buildAST(input)[0]
    input_list.append(input)
    expected_answer_list.append(expected_answer)
    result_list.append(checkAnswer(output, expected_answer))

    # q(X,Y) :- s(_,Y), t(X,_), u(_), v(_,_)

    input = ([[['q', ['X', 'Y']], ':-', ['s', ['_', 'Y']], ['t', ['X', '_']], ['u', ['_']], ['v', ['_', '_']]]])
    expected_answer = ConjunctiveQuery([RelationInQuery('s', [Wildcard(),Variable('Y')], False),
                                        RelationInQuery('t', [Variable('X'), Wildcard()], False),
                                        RelationInQuery('u', [Wildcard()], False),
                                        RelationInQuery('v', [Wildcard(),Wildcard()], False)],[],
                                       RelationInQuery('q', [Variable('X'), Variable('Y')]))
    output = ast_builder.buildAST(input)[0]
    input_list.append(input)
    expected_answer_list.append(expected_answer)
    result_list.append(checkAnswer(output, expected_answer))

    # answer(X,Y):-S(X,Z),S(Y,Z),X>Y

    input = ([[['answer', ['X', 'Y']], ':-', ['s', ['X', 'Z']], ['s', ['Y', 'Z']], ['X', '>', 'Y']]])
    expected_answer = ConjunctiveQuery([RelationInQuery('s', [Variable('X'),Variable('Z')],False),
                                        RelationInQuery('s', [Variable('Y'),Variable('Z')],False)],
                                       [Constraint(Variable('X'), Variable('Y'),'>')],
                                       RelationInQuery('answer', [Variable('X'),Variable('Y')], False))
    output = ast_builder.buildAST(input)[0]
    input_list.append(input)
    expected_answer_list.append(expected_answer)
    result_list.append(checkAnswer(output, expected_answer))

    # ---------------------------------------------------------------------------------------------------------------- #
    # ---------------------------------------------------------------------------------------------------------------- #
    # ---------------------------------------------------------------------------------------------------------------- #

    failed_test_list = []

    for i in range(0, len(result_list)):
        result = result_list[i]
        if not result:
            failed_test_list.append(i+1)

    test_score = str(len(result_list) - len(failed_test_list)) + '/' + str(len(result_list))

    print('*** Test Results ***', '\n',
          '*** Score: ', test_score, '\n',
          #'*** Inputs: ', '\n',
          #input_list, '\n',
          #"\n".join(item[0] for item in input_list), '\n',
          #'*** Expected answers: ', '\n',
          #expected_answer_list, '\n'
          #"\n".join(item[0] for item in expected_answer_list), '\n',
          '*** Failed tests: ', failed_test_list)


def checkAnswer(answer_1, answer_2):
    return str(answer_1) == str(answer_2)

def main():
    hardCodedTest()

main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Matthew Saltz'
__email__ = 'saltzm@gmail.com'
__version__ = '0.1.0'
__all__ = ["interpreter", "dataStructures", "queryExecutor", "ASTFactory", "TranslationWrapper"]



from pyparsing import *

# TODO: 1. Implement nested joins.
# TODO: 2. Look at LJ, RJ, FJ grouping. Get to some agreement on this.
# TODO: 3. Implement Datalog function (top, etc.) checking to be the FIRST thing to be evaluated in rule expression.

# Notes:
# - We decided not to implement compound terms (arithmetic expressions) for now. The definition for compound term can
# be found below under the definition of "Term".

# Definitions:
# Number:           Signed integers (no positive sign). Float with a dot between two digits. Scientific notation
#                   is supported as aEb where a = fractional number, b = integer which may start with + or -.
# Constant:         A number, any sequence of alphanumerics including underscore but starting with lowercase letter,
#                   or any sequence of characters delimited by single quotes.
# Variable:         Starts with uppercase or underscore, made of alphanumeric characters
# Unknown:          Null values represented with "null" for normal users or "'$NULL'(ID)" for development purposes,
#                   where ID is an integer.
# Term:             Noncompound: variables or constants
#                   Compound: Form of t(t1,...,tn) where t is the functor (which follows the syntax of a noncompound
#                   term) and tn are noncompound terms. In the context of YADI, it refers to arithmetic expressions.
# Predicate symbol: Defined as a sequence of alphanumerics characters + underscore that start with lowercase or
#                   underscore. On a(t1,t2), "a" stands for the predicate symbol. Relation is a synonym of predicate.
# Atom:             Has the form of a(t1, t2) for ti (0<=1<=n). If i = 0, then it is simply written as "a."
# Condition:        Boolean expression containing comparison operators, constants and variables.
# Literal:          A literal can be:
#                   -Positive (an atom).
#                   -Negative. not(body) where body is a body. Used to express the negation of a relation.
# Relation func.:   Built-in functions of the form f(a1,...,an) where ai is a relation (predicate). Built-in
#                   functions implemented are:
#                   -not(a)
#                   -lj(a1,a2,a3) is left outer join where a1 = left relation, a2 = right relation, a3 = join condition.
#                    Same is applicable for rj(a1,a2,a3) for right outer join, and fj(a1,a2,a3) for full outer join.
#                   Outer join functions can be nested.
# Head:             A positive atom that is not a built-in predicate symbol.
# Body:             Comma separated sequence of literals, which may contain built in functions, disjunctions and
#                   division.
# Rule:             head :- body, or just head (called a fact).

class SyntaxException(Exception):
    pass

class parser:
    def __init__(self):
        pass

    def parsesentence(self, sentence):
        #Reserved words
        reserved_null = Literal("is_null") | Literal("is_not_null")
        reserved_is = Literal("is")
        reserved_arith = Literal("sqrt") | Literal("log") | Literal("sin") | Literal("cos") | Literal("tan") | Literal(
            "atan") | Literal("abs") | Literal("float_integer_part") | Literal("float_fractional_part") | Literal(
            "float") | Literal("sign") | Literal("truncate") | Literal("round") | Literal("floor") | Literal("ceiling")
        reserved_not = Literal("not")
        reserved_join = Literal("lj") | Literal("rj") | Literal("fj")
        reserved_aggregates = Literal("count") | Literal("sum") | Literal("avg") | Literal("min") | Literal("max")
        reserved_group = Literal("group_by")
        reserved_duplicates = Literal("distinct")
        reserved_top = Literal("top")
        reserved_answer = Literal("answer")
        reserved_word = reserved_null | reserved_is | reserved_arith | reserved_not | reserved_join | \
                        reserved_aggregates | reserved_group | reserved_duplicates | reserved_top | reserved_answer

        #Special characters
        underscore = Word("_", max=1)                   # Only a single underscore can be used for anon. var
        comma = Literal(",").suppress()                 # Not interested in commas, delete from tokens
        separator = Literal(":-")
        dot = Literal(".").suppress()

        #Comparison operators
        greater = Literal(">")
        less = Literal("<")
        equal = Literal("=")
        g_equal = Literal(">=")
        l_equal = Literal("<=")
        comp_op = g_equal | l_equal | equal | greater | less

        #Number
        number = Combine(Word('-' + nums, nums) + Optional(Literal('.') + Word(nums)) +
                         Optional(Literal('E') + Optional(Word("-+", max=1)) + Word(nums))).setName("number")

        #Constant
        constant = (number | Combine(Word(srange('[a-z]'), alphanums + "_")) |
                    QuotedString("'", unquoteResults=False)).setName("constant")

        #Variable
        variable = (Combine(Optional("_") + Word(srange('[A-Z]')) + Optional(Word(alphanums))) |
                    Combine(Literal("_") + Word(srange('[a-z]'))) |
                    underscore).setName("variable")

        #Unknown
        unknown = (Literal("null") | Combine((Literal("'$NULL'(") + Word(nums) + Literal(")")))).setName("unknown")

        #Terms
        noncompound = (variable | constant).setName("noncompound")
        term = noncompound

        #Atom
        predicate_symbol = Combine((Optional(Literal('_')) + Word(srange('[a-z]')) + Optional(Word(alphanums + '_'))) |
                                   (OneOrMore(Literal('_')) + Word(alphanums, min=1) + Optional(
                                       Word(alphanums + '_')))).setName("predicate_symbol")

        atom = ((predicate_symbol + Literal("(").suppress() + Group(term + Optional(OneOrMore(comma | term))) +
                 Literal(")").suppress()) | predicate_symbol).setName("atom")

        #Comparison and conditions
        comparison = Group(noncompound + comp_op + noncompound).setName("comparison")

        #Literals
        conj_disj_div = Literal(",").suppress() | Literal(";") | Literal("division")
        positive = Group(atom).setName("positive_atom")
        literal = positive.setName("literal")

        #Relation functions
        not_function = Group(Literal("not") + Literal("(").suppress() + positive + Literal(")").suppress()).setName(
            "not_function")
        join_types = (Literal("lj") | Literal("rj") | Literal("fj")).setName("join_type")
        join_base = (join_types + Literal('(').suppress() + Group(atom) + comma + Group(
            atom) + comma + comparison + Literal(')').suppress()).setName("join_base")
        relation_function = (not_function | join_base).setName("relation_function")

        #Rules and facts
        head = NotAny(reserved_word) + positive.setName("head")
        body_part = relation_function | comparison | literal
        body = (body_part + ZeroOrMore(conj_disj_div + body_part)).setName("body")
        temporary_view = body
        rule = (OneOrMore(Group(((head + separator + body) | temporary_view | head) + dot)) + StringEnd()).setName(
            "expression").setFailAction(self.syntax)

        test = rule.parseString(sentence)
        return test

    def syntax(self, s, loc, expr, err):
        x = "Syntax error on string {0!r}, loc {1!r} of '{2!r}'. Details: {3!r}".format(s, loc, expr, err)
        raise SyntaxException(x)


from ..ASTFactory.ASTBuilder import *

from ..ASTFactory.ASTBuilder import ASTBuilder
from ..parserAST.Parser import parser,SyntaxException
from ..queryExecutor.queryExecutor import QueryExecutor
from ..queryExecutor.exceptions import *

__author__ = 'caioseguin'

def translateDatalogToSql(datalog_statement):

    datalog_parser = parser()
    ast_builder = ASTBuilder()
    query_executor = QueryExecutor()
    sql_query_list = []

#    print ('-----------------------------')
#    print('Datalog Input:')
#    print(datalog_statement)
#    print('\n')

    try:
        parsed_statement = datalog_parser.parsesentence(datalog_statement)
        ast_query_list = ast_builder.buildAST(parsed_statement.asList())
#        print('Read as :')

        for ast_query in ast_query_list:
            sql_query = QueryExecutor().execute_query(ast_query)
            sql_query_list.append(sql_query)
#            print('Generated SQL:')
#            print (sql_query)
    except SyntaxException as e:
        print ('SyntaxException: ' + str(e))
    except SafetyException as e:
        print ('SafetyException: ' + str(e))
    except Exception as e:
        import traceback
        traceback.print_exc()


    return sql_query_list



import copy
from ..dataStructures.query import *
from ..dataStructures.relation import *
from ..dataStructures.constraint import *
from ..queryExecutor.exceptions import *

class QueryPreprocessor():
    def preprocess(self,query):
        if isinstance(query, ConjunctiveQuery):
            pp = ConjunctiveQueryPreprocessor()
        if isinstance(query, DisjunctiveQuery):
            pp = DisjunctiveQueryPreprocessor()
        return pp.preprocess(query)

class ConjunctiveQueryPreprocessor():
    default_head_relation_name = 'answer'

    def preprocess(self,q):
        query = copy.deepcopy(q)
        var_dict = query.get_var_dict()
        query = self.create_head_variable_if_none_exists(query,var_dict)
        query = self.reduce_equality_constraints(query)
        return query

    def create_head_variable_if_none_exists(self,query,var_dict):
        if query.get_head_relation() is None:
            safe_variables = self.get_safe_variables(query,var_dict)
            query.set_head_relation(RelationInQuery(self.default_head_relation_name,safe_variables))
        return query

    ''' This method takes a query, makes a deep copy of it and modifies it to keep only equality constraints of the form:
        Var = const if Var occurs in the head OR
        const = const
        ie, constraints of the form Var1 = Var2, const = Var, Var = Const (if Var does not occur in the head) are eliminated.

        This makes several steps easier later on.
        For instance, when checking if a variable is safe.

        In the following query Y should be safe.

            R(X) := not T(Y), X = Y, X = 2

        And in the following query X should be safe

            R(X) := not T(Y), X = Y, S(Y)

        The only way to decide so is by noticing that because X=Y and X=2, Y must also be 2.
        This can be done if the equalities are reduced according to what was said previously:

            R(X) := not T(Y), X = Y, X = 2 is equivalent to:
            R(X) := not T(X), X = 2 is equivalent to:
            R(Y) := not T(Y), Y = 2 is equivalent to:

    '''
    def reduce_equality_constraints(self,query):
        # Build the equivalence sets

        equality_constraints = [x for x in query.constraints if x.is_equality_constraint()]
        eq_sets = []
        for eq_constraint in equality_constraints:
            el1 = eq_constraint.get_left_side()
            el2 = eq_constraint.get_right_side()

            set1 = -1
            set2 = -1

            for i in range(0,len(eq_sets)):
                if el1 in eq_sets[i]:
                    set1 = i
                if el2 in eq_sets[i]:
                    set2 = i

            if set1== -1 and set2== -1: # Neither El1 nor El2 is in a set.
                eq_sets.append({el1,el2})
            else: # set1 >= 0 OR set2>=0
                if set1>= 0:
                    if set2 == -1: # El1 is in some set and El2 is not.
                        eq_sets[set1].add(el2)
                    else: # El1 and El2 are both in some sets.
                        if set1!=set2: # If they're not in the same set, merge them and delete one of them.
                            eq_sets[set1] = eq_sets[set1] | eq_sets[set2]
                            eq_sets.remove(eq_sets[set2])
                else:
                    if set2 >= 0: # El2 is in a set, El1 is not.
                        eq_sets[set2].add(el1)

        new_eq_constraints = []

        for s in eq_sets:
            constants = [x for x in s if x.is_constant()]
            variables = [x for x in s if x.is_variable()]
            vars_in_rels = set([y for x in query.relations for y in x.variables if y in variables])
            # Need variables in all relations, as well as vars in equality constraints.
            # Unioning with get_safe_variables() includes those variables in equality constraints
            variables_occur_relation = list(vars_in_rels | set(self.get_safe_variables(query, query.get_var_dict())))

            if len(variables_occur_relation)==0 and len(variables)>0 :
                raise NotInstantiatedException('Equivalence set of ' + str(variables[0]) + ' does not unify with a variable in a positive goal' )

            # Substitute every variable in the equivalence set b
            if len(constants)>0:
                # Substitute every occurence of variable in every goal with the constant
                constant = constants[0]
                for relation in query.relations:
                    del_list = []

                    for variable in relation.variables:
                        if variable in s:
                            if constant in relation.constants:
                                relation.constants[constant] += relation.variables[variable]
                            else:
                                relation.constants[constant] = relation.variables[variable]
                            del_list.append(variable)
                    for variable in del_list:
                        del relation.variables[variable]

                # Substitute every occurence of the variable in the constraints with the constant
                for constraint in query.constraints:
                    el0 = constraint.get_left_side()
                    el1 = constraint.get_right_side()
                    if el0.is_variable() and el0 in s:
                        constraint.set_left_side(constant)
                    if el1.is_variable() and el1 in s:
                        constraint.set_right_side(constant)
                # Make sure we can still unify head variables with constants if they are in the equivalence set.
                for variable in variables:
                    if variable in query.head_relation.get_variables():
                        new_eq_constraints.append(Constraint(variable,constant,Constraint.EQ_OPERATOR))
            else:
                # Substitute every occurence of variable in the relations with one variable from variables_occur_relation
                var = variables_occur_relation[0]
                for relation in query.relations:
                    del_list = []
                    add_dict = {}
                    for variable in relation.variables:
                        if variable in s and (not variable == var):
                            if var in relation.variables:
                                relation.variables[var] += relation.variables[variable]
                            else:
                                add_dict[var] = relation.variables[variable]
                            del_list.append(variable)
                    for variable in del_list:
                        del relation.variables[variable]

                    for key in add_dict:
                        relation.variables[key] = add_dict[key]
                # Substitute every occurence of variable in the head with one variable from variables_occur_relation

                head_var_list = list(query.head_relation.get_variables())
                variables_in_head = query.head_relation.get_variables()
                if not var in variables_in_head:
                    variables_in_head[var] = []

                for head_var in head_var_list:
                    if head_var in s and not var == head_var:
                        variables_in_head[var] += variables_in_head.pop(head_var)

                # Substitute every occurence of variable in the constraints with one variable from variables_occur_relation
                for constraint in query.constraints:
                    el0 = constraint.get_left_side()
                    el1 = constraint.get_right_side()
                    if el0.is_variable() and el0 in s:
                        constraint.set_left_side(var)
                    if el1.is_variable() and el1 in s:
                        constraint.set_right_side(var)

            # If there's more than one constant in this equivalence set, they are all be different so we need to keep one of these constraints so the query returns false.
            if len(constants)>1:
                new_eq_constraints.append(Constraint(constants[0],constants[1],Constraint.EQ_OPERATOR))

        query.constraints = [x for x in query.constraints if not x.is_equality_constraint()] + new_eq_constraints
        return query

    def get_safe_variables(self,query,var_dict):
        # Create a list of safe variables: variables bounded to constants or that occur in a positive goal

        variables_bounded_to_constants = []

        for constraint in query.constraints:
            if (constraint.get_left_side().is_variable() and
                constraint.get_right_side().is_constant() and
                constraint.is_equality_constraint()):
                    variables_bounded_to_constants.append(constraint.get_left_side())

        safe_variables = list(var_dict) + variables_bounded_to_constants

        return safe_variables



import cmd, sys
from yadi import *
from sqlalchemy import create_engine
from sqlalchemy.engine.url import URL
from tabulate import tabulate
from ..TranslationWrapper import translateDatalogToSQL


class Interpreter(cmd.Cmd):
    intro = '\nWelcome to the YADI project.\nType help or ? to list commands.\n'
    prompt = 'YADI> '
    engine = None
    db = ""
    user = ""
    host = ""
    port  = ""

    def emptyline(self):
        return

    def parseline(self, line):
        """Parse the line into a command name and a string containing
        the arguments.  Returns a tuple containing (command, args, line).
        'command' and 'args' may be None if the line couldn't be parsed.
        """

        line = line.strip()
        if not line:
            return None, None, line
        elif line[0] == '?':
            line = 'help ' + line[1:]
            return super().parseline(line)
        elif line[0] == '/':
            return super().parseline(line[1:])
        else:
            return None, None, line

        #--commands--
    def do_open_db(self, arg):
        'Open database connection with the supplied arguments:\n\nopen_db username password database host=localhost port=5432\n'
        args = arg.split()
        args_len = len(args)

        if (args_len < 3):
            print("YADI was not able to create a database connection")
            super().do_help("open_db")
            return

        username = args[0]
        password = args[1]
        database = args[2]
        host = None
        port = None

        if(args_len >= 4):
            host  = args[3]

        if(args_len == 5):
            port = args[4]

        connection_string = URL('postgresql', username, password, host, port, database)
        new_engine = create_engine(connection_string, echo=False)

        if(new_engine):
            self.engine = new_engine
            try:
                connection = self.engine.connect()
                connection.close()
                self.user = username
                self.db = database = args[2]
                if(host):
                    self.host = host
                else:
                    self.host = "localhost"
                if(port):
                    self.port = port
                else:
                    self.port = "5432"
                print("YADI is connected to database: "+ database)
            except Exception as e:
                print("Error creating connection to database:\n{0}".format(e.args[0]))
                return
        else:
            print("YADI was not able to create a database connection")
            super().do_help("open_db")
            return

    def do_consult(self, line):
        "Loads facts or rules from a file on absolute filePath"
        return

    def do_consult_queries(self, line):
        "Load queries from a file on absolute filePath"
        return

    def do_assertion(self, line):
        "Creates a fact or a rule"
        return

    def do_current_db(self, line):
        "Information about the current database"
        if(not(self.engine)):
            print("Not connected to any database.")
        else:
            print("Current database: " + self.db+"     User: " + self.user +"\nHost: " + self.host+"     Port: " + self.port)
        return

    def do_quit(self, arg):
        'Close the YADI window, and exit'
        print('Thank you for using YADI')
        self.close()
        return True

    def default(self, line):

        if(not(self.engine)):
            print("YADI is not connected to a database. Please use /open_db command")
            super().do_help("open_db")
            return

        try:
            datalog_statement = self.preprocess(line)
            sql = translateDatalogToSql(datalog_statement)

            connection = self.engine.connect()

            for query in sql:
                print(query + '\n')
                result = connection.execute(query)
                print(tabulate(result))

            connection.close()

        except Exception as e:
            print("Error querying the database:")
            print(e)
            return

    def preprocess(self, line):
        #line = line.lower()
        #if self.file and 'playback' not in line:
        #    print(line, file=self.file)
        return line

    def close(self):

        if self.engine:
            self.engine.dispose()
            self.engine= None



__author__ = 'saltzm'

from sqlalchemy import *
from tabulate import tabulate
import os
engine = create_engine('postgresql://yadi_user:@localhost:5432/twitter', echo=True)


def create_db():
    with open(os.path.join(os.path.dirname(__file__), 'database_creation/setup.sql'), 'r') as content_file:
        content = content_file.read()
    r = engine.execute(content)
    fetched_results = r.fetchall()
    print (tabulate(fetched_results, tablefmt="grid"))
create_db()



__author__ = 'caioseguin'


from ..dataStructures.constraint import Constraint
from ..dataStructures.element import Variable, Constant, Wildcard
from ..dataStructures.query import ConjunctiveQuery
from ..dataStructures.relation import RelationInQuery

# Operators used in this module

ruleOperator = ':-'
disjunctionOperator = ';'
greaterOperator = '>'
lessOperator = '<'
greaterEqualOperator = '>='
lessEqualOperator = '<='
equalOperator = '='
negationOperator = 'not'
forbiddenSymbol = '$'

class RuleHandler:
    def __init__(self):
        pass

    def handleFact(self, fact):
        relation_object_list, constraint_object_list = self.handleBody(fact)
        return ConjunctiveQuery(relation_object_list, constraint_object_list, None)

    def handleConjunctiveRule(self, rule):

        assert isinstance(rule, list)

        #if len(rule) != 1:
        #    raise Exception("list with exactly one element expected")
        #else:
        #    rule = rule[0] # Peal list

        head = self.extractHead(rule)
        body = self.extractBody(rule)

        relation_object_list, constraint_object_list = self.handleBody(body)
        query_object = self.handleHead(head, relation_object_list, constraint_object_list)

        return query_object

    def handleHead(self, head, relation_object_list, constraint_object_list):
        head_relation = self.handleRelation(head)
        return ConjunctiveQuery(relation_object_list, constraint_object_list, head_relation)

    def handleBody(self, body):

        assert isinstance(body, list)

        relation_list = []
        constraint_list = []

        for body_part in body:
            if self.isConstraint(body_part):
                constraint_list.append(self.handleConstraint(body_part))
            else:
                relation_list.append(self.handleRelation(body_part))

        return relation_list, constraint_list

    def handleRelation(self, relation):

        assert isinstance(relation, list)

        if self.isNegatedRelation(relation):
            is_negated = True
            relation = relation[1:len(relation)][0]
        else:
            is_negated = False

        relation_symbol = relation[0]
        # TODO
#        if len(relation) == 1:
#            raise Exception("Atoms not yet supported")

        if not isinstance(relation[1], list):
            raise Exception("list of terms expected")
        else:
            term_list = relation[1]

        ast_term_list = self.handleTermList(term_list)
        new_relation_in_query = RelationInQuery(relation_symbol, ast_term_list, is_negated)
        return new_relation_in_query


    # Will have to be updated to deal with strings as constants
    def handleTermList(self, term_list):

        ast_term_list = []

        for term in term_list:
            if self.isVariable(term):
                ast_term_list.append(Variable(term))
            if self.isConstant(term):
                ast_term_list.append(Constant(self.handleConstant(term)))
            if self.isWildcard(term):
                ast_term_list.append(Wildcard())

        return ast_term_list

    # Will have to be updated to deal with strings as constants
    def handleConstraint(self, constraint):

        if self.isVariable(constraint[0]):
            left_side = Variable(constraint[0])
        else:
            left_side = Constant(self.handleConstant(constraint[0]))

        if self.isVariable(constraint[2]):
            right_side = Variable(constraint[2])
        else:
            right_side = Constant(self.handleConstant(constraint[2]))

        return Constraint(left_side, right_side, constraint[1])

    def extractHead(self, rule):
        return rule[0]

    def extractBody(self, rule):
        if not len(rule) >= 3:
            raise Exception("list with three or more elements expected for a body")
        return rule[2:len(rule)]

    def isRule(self, statement):
        return ruleOperator in statement

    def isDisjunctiveRule(self, statement):
        return disjunctionOperator in statement[0]

    def isNegatedRelation(self, relation):
        return relation[0] == negationOperator

    def isConstraint(self, body_part):
        return (len(body_part) == 3) and ( ( body_part[1] == greaterOperator ) or
                                           ( body_part[1] == greaterEqualOperator ) or
                                           ( body_part[1] == lessOperator ) or
                                           ( body_part[1] == lessEqualOperator ) or
                                           ( body_part[1] == equalOperator ))

    def isVariable(self, term):
        return term.isupper()

    def isConstant(self, term):
        return term.islower() or term.isdigit() or (term[0] == "'" and term[len(term)-1] == "'")

    def isWildcard(self, term):
        return term == '_'

    def handleConstant(self, constant):
        if constant[0] == "'" and constant[len(constant)-1] == "":
            return constant[1:len(constant)-1]
        else:
            return constant



import sqlalchemy


#from evaluate_query import *
from .dataStructures.query import *
from .dataStructures.relation import *
from .dataStructures.element import *
from .dataStructures.constraint import *
from .queryExecutor.queryExecutor import *
from .queryExecutor.sqlFactory import *


def test(list_queries):
    gen = SQLGenerator()
    for i in range(0,len(list_queries)):
        print ('Test :'+ str(i))
        print ('Original:')
        print (str(list_queries[i]))
        print ('Result:')
        print (QueryExecutor().execute_query(list_queries[i]))
        print ('---------------------------------------------------------')



queries = []

#M(title) :- movie(title,length_min), length_mins>100.
r = RelationInQuery('movie', [Variable('title'),Wildcard(),Variable('length_mins'), Wildcard()], False)
q = ConjunctiveQuery([r], [Constraint(Variable('length_mins'), Constant('100'), '>=')], RelationInQuery('M', [Variable('title')]))

queries.append(q)
# ------
# Q(X):-!S(Y),X=Y,Y=2
r = RelationInQuery('S', [Variable('Y')],True)
q = ConjunctiveQuery([r],[Constraint(Variable('X'), Variable('Y'), '='),Constraint(Variable('Y'), Constant('2'), '=')],RelationInQuery('Q', [Variable('X')]))

queries.append(q)

#------
#Q(X):- S(X)
r = RelationInQuery('S', [Variable('X')])
q = ConjunctiveQuery([r],[],RelationInQuery('Q', [Variable('X')]))
queries.append(q)

# ------
# Q(X):- S(X), X = 2
r = RelationInQuery('S', [Variable('X')])
head = RelationInQuery('Q', [Variable('X')])
queries.append(ConjunctiveQuery([r],[Constraint(Variable('X'),Constant('2'),'=')],head))

# ------

s = RelationInQuery('S', [Variable('X'),Variable('Y')])
t = RelationInQuery('T', [Variable('X')], True)
queries.append(ConjunctiveQuery([s,t],[Constraint(Variable('Y'), Variable('Z'), '=')],RelationInQuery('Q',[Variable('X'),Variable('Z')])))


# ------

# R(X,Y),!S(Z), Y=Z
s = RelationInQuery('R', [Variable('X'),Variable('Y')])
t = RelationInQuery('S', [Variable('Z')], True)

queries.append(ConjunctiveQuery([s,t],[Constraint(Variable('Y'), Variable('Z'), '=')]))

# R(X,Y),!S(Z)

s = RelationInQuery('R', [Variable('X'),Variable('Y')])
t = RelationInQuery('S', [Variable('Z')], True)

queries.append(ConjunctiveQuery([s,t],[]))


# answer(X,Y) :- R(X,A), S(A,Y).

s = RelationInQuery('R', [Variable('X'),Variable('A')])
t = RelationInQuery('S', [Variable('A'),Variable('Y')])
head = RelationInQuery('answer',[Variable('X'),Variable('Y')])
queries.append(ConjunctiveQuery([s,t],[],head))

# answer(X,2) :- R(X,A), S(A,_).

s = RelationInQuery('R', [Variable('X'),Variable('A')])
t = RelationInQuery('S', [Variable('A'),Wildcard()])
head = RelationInQuery('answer',[Variable('X'),Constant('2')])
queries.append(ConjunctiveQuery([s,t],[],head))


# r(_,2)

r = RelationInQuery('answer',[Wildcard(),Constant('2')])
queries.append(ConjunctiveQuery([r],[]))

# R(_,A), S(A,_). -> answer(A) :- R(_,A), S(A,_).

s = RelationInQuery('R', [Wildcard(),Variable('A')])
t = RelationInQuery('S', [Variable('A'),Wildcard()])

queries.append(ConjunctiveQuery([s,t],[]))

# R(X,Y) :- S(X), S(Y), X>Y

s = RelationInQuery('S', [Variable('X')])
t = RelationInQuery('S', [Variable('Y')])
head = RelationInQuery('answer',[Variable('X'),Variable('Y')])
queries.append(ConjunctiveQuery([s,t],[Constraint(Variable('X'),Variable('Y'),'>')],head))

# R(X,Y) :- S(X), Y>2

s = RelationInQuery('S', [Variable('X')])

head = RelationInQuery('answer',[Variable('X'),Variable('Y')])
queries.append(ConjunctiveQuery([s],[Constraint(Variable('Y'),Constant('2'),'>')],head))

# R(X) :- S(X), X<2

s = RelationInQuery('S', [Variable('X')])

head = RelationInQuery('answer',[Variable('X'),Variable('Y')])
queries.append(ConjunctiveQuery([s],[Constraint(Variable('X'),Constant('2'),'<')],head))

# answer(X,Y):-S(X,Z),S(Y,Z),X>Y

s = RelationInQuery('S', [Variable('X'),Variable('Z')])
t = RelationInQuery('S', [Variable('Y'),Variable('Z')])
head = RelationInQuery('answer',[Variable('X'),Variable('Y')])

queries.append(ConjunctiveQuery([s,t],[Constraint(Variable('X'),Variable('Y'),'>')],head))


# R(X) :- X = 2, 3<X

head = RelationInQuery('R',[Variable('X')])

queries.append(ConjunctiveQuery([],[Constraint(Variable('X'),Constant('2'),'='),Constraint(Constant('3'),Variable('X'),'<')],head))

# r(_,2), X = 2

r = RelationInQuery('R',[Wildcard(),Constant('2')])
queries.append(ConjunctiveQuery([r],[Constraint(Variable('X'), Constant('2'), '=')]))

# R(X,Y) :- S(X), S(Y,T), T(X), U(Y), X>Y
s = RelationInQuery('S', [Variable('X')])
t = RelationInQuery('S', [Variable('Y')])
u = RelationInQuery('T', [Variable('X')])
v = RelationInQuery('V', [Variable('Y')])
head = RelationInQuery('answer',[Variable('X'),Variable('Y')])
queries.append(ConjunctiveQuery([s,t,u,v],[Constraint(Variable('X'),Variable('Y'),'>')],head))
test(queries)





class Constraint():
    EQ_OPERATOR = '='
    def __init__(self, left_side, right_side, operator):
        self.left_side = left_side
        self.right_side = right_side
        self.operator = operator

    def is_equality_constraint(self):
        return self.operator == Constraint.EQ_OPERATOR

    def get_left_side(self):
        return self.left_side

    def get_right_side(self):
        return self.right_side

    def get_operator(self):
        return self.operator

    def set_left_side(self,el):
        self.left_side = el

    def set_right_side(self,el):
        self.right_side = el

    def __repr__(self):
        return str(self.get_left_side()) + str(self.get_operator()) + str(self.get_right_side())




from .safetyChecker import *
from .preprocessor import *
from .sqlFactory import *

class QueryExecutor():
    def execute_query(self, query):
        print('Original query before processing: ' + str(query))
        new_query = QueryPreprocessor().preprocess(query)
        print('Info:')
        print('Query transformed into: ' + str(new_query))
        SafetyChecker().check_for_safety(new_query)

        #print ('Processing: \n' + str(new_query))
        sql = SQLGenerator().get_SQL_code(new_query,query)
        return sql



class Query():
    def get_relations(self):
        pass

class DisjunctiveQuery(Query):
    def __init__(self, queries = []):
        self.queries = queries

    def get_queries(self):
        return self.queries

    def get_relations(self):
        return [x.get_relations for x in self.get_queries()]


class ConjunctiveQuery(Query):
    def __init__(self, relations = [], constraints = [], head_relation = None):
        self.head_relation = head_relation
        self.relations = relations      # [RelationInQuery].
        self.constraints = constraints  # Explicit constraints of the form
                                        # Element COMP Element type.
                                        # [Constraint]

    def set_head_relation(self,hr):
        self.head_relation = hr

    def get_head_relation(self):
        return self.head_relation

    def get_relations(self):
        return self.relations

    def get_constraints(self):
        return self.constraints

    def __repr__(self):

        return str(self.head_relation) + ':-' + \
               ','.join([str(x) for x in self.relations]) + \
               (',' if len(self.constraints)>0 else '') + \
               ','.join([str(x) for x in self.constraints])

    # var_dict is a Variable -> [(Relation, field)] which is usefuld for several purposes.
    # It does not map occurences of variables in negated goals.
    def get_var_dict(self):
        variables_in_positive_goals = [y for x in self.relations for y in x.variables if not x.is_negated()]
        var_dict = {}

        for relation in [x for x in self.relations if not x.is_negated()]:
            for var in relation.variables:
                if not var in var_dict:
                    var_dict[var] = []
                for position in relation.variables[var]:
                    var_dict[var].append((relation,position))

        return var_dict




from .Parser import parser

p = parser()

#Tests to check syntax
print(p.parsesentence("q."))                                    # Atom, zero arity
print(p.parsesentence("q(x)."))                                 # Atom, one var
print(p.parsesentence("q('3')."))                               # Atom, string
print(p.parsesentence("q(x,y)."))                               # Atom, two-arity
print(p.parsesentence("q(_,x)."))                               # Atom, anonymous variable
print(p.parsesentence("_ab(a)."))                               # Predicate symbol with underscore
print(p.parsesentence("q2(x,z,b,'a')."))                        # Predicate symbol with number
print(p.parsesentence("__ab_55(a,b,c)."))                       # Predicate symbol with number and underscore
print(p.parsesentence("q(x,y) :- k(x,y)."))                     # Rule with one literal
print(p.parsesentence("q(x,y) :- a(foo_foo)."))                 # Rule with one literal using constant
print(p.parsesentence("q(x,y) :- k(_ab)."))                     # Rule with one literal with constant starting with underscore
print(p.parsesentence("q(x,y) :- k(X)."))                       # Rule with one literal with one variable
print(p.parsesentence("q(x,y) :- k(x,h), _v3(n,k)."))           # Rule with two literals
print(p.parsesentence("q(x,y) :- a;b."))                        # Rule with disjunction of two zero-arity atoms
print(p.parsesentence("q(x,y) :- a(x);b(x)."))                  # Rule with disjunction of two 1-arity atoms
print(p.parsesentence("q(x,y) :- a division b."))               # Rule with division of two zero-arity atoms
print(p.parsesentence("q(x,y) :- a(x,y) division b(x,y)."))     # Rule with division of two two-arity atoms
print(p.parsesentence("q(x,y,z) :- a(x),a;b."))                 # Rule with one-arity atom, disjunction of two zero-arity atoms
print(p.parsesentence("q(x,y) :- a(x), t>5."))                  # Rule with one-arity atom, boolean comparison
print(p.parsesentence("q(x,y) :- a(x), t<5."))                  # Rule with one-arity atom, boolean comparison
print(p.parsesentence("q(x,y) :- a(x), t>=5."))                 # Rule with one-arity atom, boolean comparison
print(p.parsesentence("q(x,y) :- a(x), t<=5."))                 # Rule with one-arity atom, boolean comparison
print(p.parsesentence("q(x,y) :- a(x), gd=5."))                 # Rule with one-arity atom, boolean comparison
print(p.parsesentence("q(x,y,z) :- a(x), t=4.0."))              # Rule with one-arity atom, comparison using float
print(p.parsesentence("q(x,y,z) :- a(x), t=4.0E6."))            # Rule with one-arity atom, comparison using float+E
print(p.parsesentence("q(x,y,z) :- a(x), t=4.0E+6."))           # Rule with one-arity atom, comparison using float+E+'+'
print(p.parsesentence("q(x,y,z) :- a(x), t=4.0E-6."))           # Rule with one-arity atom, comparison using float+E+'-'
print(p.parsesentence("q(x,y,z) :- a(x), t=4.0, k(x)."))        # Rule with one-arity atom, comparison, atom
print(p.parsesentence("q(x) :- x(g), not(a(x,y))."))            # Rule with one-arity atom, negation
print(p.parsesentence("q(x,y). k(x)."))                         # Two facts in a line.
print(p.parsesentence("q(x,y). q(x,y) :- a(b,c)."))             # A fact and a rule in a line.
print(p.parsesentence("q(x,y). q(x,y) :- a(b,c). a(b)."))       # A fact, a rule and a fact in a line.
print(p.parsesentence("q(x,y) :- a(b), X=3; Y>5."))             # Rule with one-arity atom, disjunctive comparison.
print(p.parsesentence("q(x,y) :- a(b), X=3, Y>5."))             # Rule with one-arity atom, conjunctive comparison.
print(p.parsesentence("q(x,y) :- a(b), X=3, Y>5, X=3; Y>5."))   # Rule with one-arity atom, two two-term comparisons.
print(p.parsesentence("r(X) :- not(t(Y)), X = Y, s(Y)."))       # Rule with a negation in front.
print(p.parsesentence("r(x) :- r(a,X); not(q(X,b)), lj(a,b,x)."))      # Rule with a left join
print(p.parsesentence("q(X,Z) :- s(X,Y), not(t(X)), Y=Z."))
print(p.parsesentence("q(X,Z) :- t>5, s(X,Y), not(t(X)), Y=Z."))
print(p.parsesentence("q(X,Y):- s(X).\nq(X,Y):- s(Y)."))        # Two statements broken down in two lines.
print(p.parsesentence("q(x,y) :- a(b), X=3, 3>Y, X=3; 5>X."))   # Rule with one-arity atom, two two-term comparisons.
print(p.parsesentence("q(X,Y), s(x)."))                         # Temporary view
print(p.parsesentence("q(X,Y), not(x(t,y))."))                  # Temporary view
print(p.parsesentence("q(X,Y):- s(X).\nq(X,Y):- s(X).\nq(X,Y):- s(X)."))
print(p.parsesentence("q(X,3) :- s(X)."))

#Incorporation of all elements
print(p.parsesentence("_a45(x,Y,_343,'a') :- __x_43A(k,3.5E+3,x), A>=4; t=5, a(q,x);r(x,Y), a division y. q(x,y)."))

#Rules (that actually make sense)
print(p.parsesentence("q(X,Y):- s(X)."))
print(p.parsesentence("q(X):- s(X)."))
print(p.parsesentence("q(X):- s(X), not(t(U))."))
print(p.parsesentence("q(X):- s(X,U), not(t(U))."))
print(p.parsesentence("q(X):- s(X), not(t(U)), U = 2."))
print(p.parsesentence("q(X):- s(X), not(t(U)), U < 2."))
print(p.parsesentence("q(X):- s(X), not(t(U)), U = X."))
print(p.parsesentence("q(X):- s(X), Y < 3."))
print(p.parsesentence("q(X):- s(X,Y), Y < 3."))
print(p.parsesentence("q(X):- s(X), not(t(Y)), X = Y."))
print(p.parsesentence("q(X,Z):- s(X,Y), not(t(A,Z)), Z = Y."))
print(p.parsesentence("q(X):- s(X), X = 2."))
print(p.parsesentence("q(X):- s(X, Y), Y = 2."))
print(p.parsesentence("q(X):- s(X, Y, Z), Y = 2, Z = Y."))
print(p.parsesentence("q(X) :- not(s(Y)),  X = 2, X = Y."))
print(p.parsesentence("q(X) :- not(s(Y)), X = Y, X = 2."))
print(p.parsesentence("q(X) :- s(X), X = Y."))
print(p.parsesentence("q(X) :- s(X), P = Y."))
print(p.parsesentence("r(X) :- s(X), 3=X, X>2."))
print(p.parsesentence("r(Y) :- s(X), Y=X, X=2, Y =4."))
print(p.parsesentence("r(X,Y,Z,_,2) :- s(X), Y=X, X=2."))
print(p.parsesentence("q(X,Y) :- s(_,Y), t(X,_), u(_), v(_,_)."))
print(p.parsesentence("q(x,y)."))
